@@ -2,63 +2,62 @@
 
 Preliminary documentation at https://github.com/ipython/ipython/wiki/IPEP-16%3A-Notebook-multi-directory-dashboard-and-URL-mapping#kernels-api
 """
-
 # Copyright (c) Jupyter Development Team.
 # Distributed under the terms of the Modified BSD License.
-
 import json
 import logging
 from textwrap import dedent
 
-from tornado import web, gen
+from ipython_genutils.py3compat import cast_unicode
+from jupyter_client import protocol_version as client_protocol_version
+
+try:
+    from jupyter_client.jsonutil import json_default
+except ImportError:
+    from jupyter_client.jsonutil import date_default as json_default
+from tornado import gen
+from tornado import web
 from tornado.concurrent import Future
 from tornado.ioloop import IOLoop
 
-from jupyter_client import protocol_version as client_protocol_version
-from jupyter_client.jsonutil import date_default
-from ipython_genutils.py3compat import cast_unicode
-from jupyter_server.utils import url_path_join, url_escape, ensure_async
-
 from ...base.handlers import APIHandler
-from ...base.zmqhandlers import AuthenticatedZMQStreamHandler, deserialize_binary_message
-
+from ...base.zmqhandlers import AuthenticatedZMQStreamHandler
+from ...base.zmqhandlers import deserialize_binary_message
+from jupyter_server.utils import ensure_async
+from jupyter_server.utils import url_escape
+from jupyter_server.utils import url_path_join
 
 
 class MainKernelHandler(APIHandler):
-
     @web.authenticated
     async def get(self):
         km = self.kernel_manager
         kernels = await ensure_async(km.list_kernels())
-        self.finish(json.dumps(kernels, default=date_default))
+        self.finish(json.dumps(kernels, default=json_default))
 
     @web.authenticated
     async def post(self):
         km = self.kernel_manager
         model = self.get_json_body()
         if model is None:
-            model = {
-                'name': km.default_kernel_name
-            }
+            model = {"name": km.default_kernel_name}
         else:
-            model.setdefault('name', km.default_kernel_name)
+            model.setdefault("name", km.default_kernel_name)
 
-        kernel_id = await km.start_kernel(kernel_name=model['name'],
-                                                       path=model.get('path'))
+        kernel_id = await km.start_kernel(kernel_name=model["name"], path=model.get("path"))
         model = await ensure_async(km.kernel_model(kernel_id))
-        location = url_path_join(self.base_url, 'api', 'kernels', url_escape(kernel_id))
-        self.set_header('Location', location)
+        location = url_path_join(self.base_url, "api", "kernels", url_escape(kernel_id))
+        self.set_header("Location", location)
         self.set_status(201)
-        self.finish(json.dumps(model, default=date_default))
+        self.finish(json.dumps(model, default=json_default))
 
 
 class KernelHandler(APIHandler):
-
     @web.authenticated
     async def get(self, kernel_id):
         km = self.kernel_manager
         model = await ensure_async(km.kernel_model(kernel_id))
-        self.finish(json.dumps(model, default=date_default))
+        self.finish(json.dumps(model, default=json_default))
 
     @web.authenticated
     async def delete(self, kernel_id):
@@ -69,14 +68,13 @@ class KernelHandler(APIHandler):
 
 
 class KernelActionHandler(APIHandler):
-
     @web.authenticated
     async def post(self, kernel_id, action):
         km = self.kernel_manager
-        if action == 'interrupt':
+        if action == "interrupt":
             await ensure_async(km.interrupt_kernel(kernel_id))
             self.set_status(204)
-        if action == 'restart':
+        if action == "restart":
 
             try:
                 await km.restart_kernel(kernel_id)
@@ -85,14 +83,14 @@ class KernelActionHandler(APIHandler):
                 self.set_status(500)
             else:
                 model = await ensure_async(km.kernel_model(kernel_id))
-                self.write(json.dumps(model, default=date_default))
+                self.write(json.dumps(model, default=json_default))
         self.finish()
 
 
 class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
-    '''There is one ZMQChannelsHandler per running kernel and it oversees all
+    """There is one ZMQChannelsHandler per running kernel and it oversees all
     the sessions.
-    '''
+    """
 
     # class-level registry of open sessions
     # allows checking for conflict on session-id,
@@ -102,28 +100,28 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
     @property
     def kernel_info_timeout(self):
         km_default = self.kernel_manager.kernel_info_timeout
-        return self.settings.get('kernel_info_timeout', km_default)
+        return self.settings.get("kernel_info_timeout", km_default)
 
     @property
     def iopub_msg_rate_limit(self):
-        return self.settings.get('iopub_msg_rate_limit', 0)
+        return self.settings.get("iopub_msg_rate_limit", 0)
 
     @property
     def iopub_data_rate_limit(self):
-        return self.settings.get('iopub_data_rate_limit', 0)
+        return self.settings.get("iopub_data_rate_limit", 0)
 
     @property
     def rate_limit_window(self):
-        return self.settings.get('rate_limit_window', 1.0)
+        return self.settings.get("rate_limit_window", 1.0)
 
     def __repr__(self):
-        return "%s(%s)" % (self.__class__.__name__, getattr(self, 'kernel_id', 'uninitialized'))
+        return "%s(%s)" % (self.__class__.__name__, getattr(self, "kernel_id", "uninitialized"))
 
     def create_stream(self):
         km = self.kernel_manager
         identity = self.session.bsession
-        for channel in ('iopub', 'shell', 'control', 'stdin'):
-            meth = getattr(km, 'connect_' + channel)
+        for channel in ("iopub", "shell", "control", "stdin"):
+            meth = getattr(km, "connect_" + channel)
             self.channels[channel] = stream = meth(self.kernel_id, identity=identity)
             stream.channel = channel
 
@@ -201,25 +199,19 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
             # NOTE: this close check appears to never be True during on_open,
             # even when the peer has closed the connection
             if self.ws_connection is None or self.ws_connection.is_closing():
-                self.log.debug(
-                    "Nudge: cancelling on closed websocket: %s", self.kernel_id
-                )
+                self.log.debug("Nudge: cancelling on closed websocket: %s", self.kernel_id)
                 finish()
                 return
 
             # check for stopped kernel
             if self.kernel_id not in self.kernel_manager:
-                self.log.debug(
-                    "Nudge: cancelling on stopped kernel: %s", self.kernel_id
-                )
+                self.log.debug("Nudge: cancelling on stopped kernel: %s", self.kernel_id)
                 finish()
                 return
 
             # check for closed zmq socket
             if shell_channel.closed():
-                self.log.debug(
-                    "Nudge: cancelling on closed zmq socket: %s", self.kernel_id
-                )
+                self.log.debug("Nudge: cancelling on closed zmq socket: %s", self.kernel_id)
                 finish()
                 return
 
@@ -266,7 +258,7 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
 
         enabling msg spec adaptation, if necessary
         """
-        idents,msg = self.session.feed_identities(msg)
+        idents, msg = self.session.feed_identities(msg)
         try:
             msg = self.session.deserialize(msg)
         except:
@@ -274,9 +266,9 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
             self._kernel_info_future.set_result({})
             return
         else:
-            info = msg['content']
+            info = msg["content"]
             self.log.debug("Received kernel info: %s", info)
-            if msg['msg_type'] != 'kernel_info_reply' or 'protocol_version' not in info:
+            if msg["msg_type"] != "kernel_info_reply" or "protocol_version" not in info:
                 self.log.error("Kernel info request failed, assuming current %s", info)
                 info = {}
             self._finish_kernel_info(info)
@@ -292,10 +284,16 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
         Set up protocol adaptation, if needed,
         and signal that connection can continue.
         """
-        protocol_version = info.get('protocol_version', client_protocol_version)
+        protocol_version = info.get("protocol_version", client_protocol_version)
         if protocol_version != client_protocol_version:
-            self.session.adapt_version = int(protocol_version.split('.')[0])
-            self.log.info("Adapting from protocol version {protocol_version} (kernel {kernel_id}) to {client_protocol_version} (client).".format(protocol_version=protocol_version, kernel_id=self.kernel_id, client_protocol_version=client_protocol_version))
+            self.session.adapt_version = int(protocol_version.split(".")[0])
+            self.log.info(
+                "Adapting from protocol version {protocol_version} (kernel {kernel_id}) to {client_protocol_version} (client).".format(
+                    protocol_version=protocol_version,
+                    kernel_id=self.kernel_id,
+                    client_protocol_version=client_protocol_version,
+                )
+            )
         if not self._kernel_info_future.done():
             self._kernel_info_future.set_result(info)
 
@@ -307,7 +305,7 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
         self.kernel_info_channel = None
         self._kernel_info_future = Future()
         self._close_future = Future()
-        self.session_key = ''
+        self.session_key = ""
 
         # Rate limiting code
         self._iopub_window_msg_count = 0
@@ -337,13 +335,14 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
                 return
             self.log.warning("Timeout waiting for kernel_info reply from %s", self.kernel_id)
             future.set_result({})
+
         loop = IOLoop.current()
         loop.add_timeout(loop.time() + self.kernel_info_timeout, give_up)
         # actually wait for it
         await future
 
     async def get(self, kernel_id):
-        self.kernel_id = cast_unicode(kernel_id, 'ascii')
+        self.kernel_id = cast_unicode(kernel_id, "ascii")
         await super(ZMQChannelsHandler, self).get(kernel_id=kernel_id)
 
     async def _register_session(self):
@@ -353,7 +352,7 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
         This is likely due to a client reconnecting from a lost network connection,
         where the socket on our side has not been cleaned up yet.
         """
-        self.session_key = '%s:%s' % (self.kernel_id, self.session.session)
+        self.session_key = "%s:%s" % (self.kernel_id, self.session.session)
         stale_handler = self._open_sessions.get(self.session_key)
         if stale_handler:
             self.log.warning("Replacing stale connection: %s", self.session_key)
@@ -367,7 +366,7 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
 
         # on new connections, flush the message buffer
         buffer_info = km.get_buffer(kernel_id, self.session_key)
-        if buffer_info and buffer_info['session_key'] == self.session_key:
+        if buffer_info and buffer_info["session_key"] == self.session_key:
             self.log.info("Restoring connection for %s", self.session_key)
             if km.ports_changed(kernel_id):
                 # If the kernel's ports have changed (some restarts trigger this)
@@ -375,12 +374,12 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
                 self.create_stream()
             else:
                 # The kernel's ports have not changed; use the channels captured in the buffer
-                self.channels = buffer_info['channels']
+                self.channels = buffer_info["channels"]
 
             connected = self.nudge()
 
             def replay(value):
-                replay_buffer = buffer_info['buffer']
+                replay_buffer = buffer_info["buffer"]
                 if replay_buffer:
                     self.log.info("Replaying %s buffered messages", len(replay_buffer))
                     for channel, msg_list in replay_buffer:
@@ -403,7 +402,7 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
                 return
 
         km.add_restart_callback(self.kernel_id, self.on_kernel_restarted)
-        km.add_restart_callback(self.kernel_id, self.on_restart_failed, 'dead')
+        km.add_restart_callback(self.kernel_id, self.on_restart_failed, "dead")
 
         def subscribe(value):
             for channel, stream in self.channels.items():
@@ -413,7 +412,6 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
 
         return connected
 
-
     def on_message(self, msg):
         if not self.channels:
             # already closed, ignore the message
@@ -423,15 +421,15 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
             msg = deserialize_binary_message(msg)
         else:
             msg = json.loads(msg)
-        channel = msg.pop('channel', None)
+        channel = msg.pop("channel", None)
         if channel is None:
             self.log.warning("No channel specified, assuming shell: %s", msg)
-            channel = 'shell'
+            channel = "shell"
         if channel not in self.channels:
             self.log.warning("No such channel: %r", channel)
             return
         am = self.kernel_manager.allowed_message_types
-        mt = msg['header']['msg_type']
+        mt = msg["header"]["msg_type"]
         if am and mt not in am:
             self.log.warning('Received message of type "%s", which is not allowed. Ignoring.' % mt)
         else:
@@ -441,23 +439,27 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
     def _on_zmq_reply(self, stream, msg_list):
         idents, fed_msg_list = self.session.feed_identities(msg_list)
         msg = self.session.deserialize(fed_msg_list)
-        parent = msg['parent_header']
+        parent = msg["parent_header"]
+
         def write_stderr(error_message):
             self.log.warning(error_message)
-            msg = self.session.msg("stream",
-                content={"text": error_message + '\n', "name": "stderr"},
-                parent=parent
+            msg = self.session.msg(
+                "stream", content={"text": error_message + "\n", "name": "stderr"}, parent=parent
             )
-            msg['channel'] = 'iopub'
-            self.write_message(json.dumps(msg, default=date_default))
-        channel = getattr(stream, 'channel', None)
-        msg_type = msg['header']['msg_type']
+            msg["channel"] = "iopub"
+            self.write_message(json.dumps(msg, default=json_default))
 
-        if channel == 'iopub' and msg_type == 'error':
+        channel = getattr(stream, "channel", None)
+        msg_type = msg["header"]["msg_type"]
+
+        if channel == "iopub" and msg_type == "error":
             self._on_error(msg)
 
-
-        if channel == 'iopub' and msg_type == 'status' and msg['content'].get('execution_state') == 'idle':
+        if (
+            channel == "iopub"
+            and msg_type == "status"
+            and msg["content"].get("execution_state") == "idle"
+        ):
             # reset rate limit counter on status=idle,
             # to avoid 'Run All' hitting limits prematurely.
             self._iopub_window_byte_queue = []
@@ -466,13 +468,13 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
             self._iopub_msgs_exceeded = False
             self._iopub_data_exceeded = False
 
-        if channel == 'iopub' and msg_type not in {'status', 'comm_open', 'execute_input'}:
+        if channel == "iopub" and msg_type not in {"status", "comm_open", "execute_input"}:
 
             # Remove the counts queued for removal.
             now = IOLoop.current().time()
             while len(self._iopub_window_byte_queue) > 0:
                 queued = self._iopub_window_byte_queue[0]
-                if (now >= queued[0]):
+                if now >= queued[0]:
                     self._iopub_window_byte_count -= queued[1]
                     self._iopub_window_msg_count -= 1
                     del self._iopub_window_byte_queue[0]
@@ -483,7 +485,7 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
 
             # Increment the bytes and message count
             self._iopub_window_msg_count += 1
-            if msg_type == 'stream':
+            if msg_type == "stream":
                 byte_count = sum([len(x) for x in msg_list])
             else:
                 byte_count = 0
@@ -502,7 +504,9 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
             if self.iopub_msg_rate_limit > 0 and msg_rate > self.iopub_msg_rate_limit:
                 if not self._iopub_msgs_exceeded:
                     self._iopub_msgs_exceeded = True
-                    write_stderr(dedent("""\
+                    write_stderr(
+                        dedent(
+                            """\
                     IOPub message rate exceeded.
                     The Jupyter server will temporarily stop sending output
                     to the client in order to avoid crashing it.
@@ -512,7 +516,11 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
                     Current values:
                     ServerApp.iopub_msg_rate_limit={} (msgs/sec)
                     ServerApp.rate_limit_window={} (secs)
-                    """.format(self.iopub_msg_rate_limit, self.rate_limit_window)))
+                    """.format(
+                                self.iopub_msg_rate_limit, self.rate_limit_window
+                            )
+                        )
+                    )
             else:
                 # resume once we've got some headroom below the limit
                 if self._iopub_msgs_exceeded and msg_rate < (0.8 * self.iopub_msg_rate_limit):
@@ -524,7 +532,9 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
             if self.iopub_data_rate_limit > 0 and data_rate > self.iopub_data_rate_limit:
                 if not self._iopub_data_exceeded:
                     self._iopub_data_exceeded = True
-                    write_stderr(dedent("""\
+                    write_stderr(
+                        dedent(
+                            """\
                     IOPub data rate exceeded.
                     The Jupyter server will temporarily stop sending output
                     to the client in order to avoid crashing it.
@@ -534,7 +544,11 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
                     Current values:
                     ServerApp.iopub_data_rate_limit={} (bytes/sec)
                     ServerApp.rate_limit_window={} (secs)
-                    """.format(self.iopub_data_rate_limit, self.rate_limit_window)))
+                    """.format(
+                                self.iopub_data_rate_limit, self.rate_limit_window
+                            )
+                        )
+                    )
             else:
                 # resume once we've got some headroom below the limit
                 if self._iopub_data_exceeded and data_rate < (0.8 * self.iopub_data_rate_limit):
@@ -565,10 +579,13 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
         if self.kernel_id in km:
             km.notify_disconnect(self.kernel_id)
             km.remove_restart_callback(
-                self.kernel_id, self.on_kernel_restarted,
+                self.kernel_id,
+                self.on_kernel_restarted,
             )
             km.remove_restart_callback(
-                self.kernel_id, self.on_restart_failed, 'dead',
+                self.kernel_id,
+                self.on_restart_failed,
+                "dead",
             )
 
             # start buffering instead of closing if this was the last connection
@@ -589,36 +606,35 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
         self._close_future.set_result(None)
 
     def _send_status_message(self, status):
-        iopub = self.channels.get('iopub', None)
+        iopub = self.channels.get("iopub", None)
         if iopub and not iopub.closed():
             # flush IOPub before sending a restarting/dead status message
             # ensures proper ordering on the IOPub channel
             # that all messages from the stopped kernel have been delivered
             iopub.flush()
-        msg = self.session.msg("status",
-            {'execution_state': status}
-        )
-        msg['channel'] = 'iopub'
-        self.write_message(json.dumps(msg, default=date_default))
+        msg = self.session.msg("status", {"execution_state": status})
+        msg["channel"] = "iopub"
+        self.write_message(json.dumps(msg, default=json_default))
 
     def on_kernel_restarted(self):
         logging.warn("kernel %s restarted", self.kernel_id)
-        self._send_status_message('restarting')
+        self._send_status_message("restarting")
 
     def on_restart_failed(self):
         logging.error("kernel %s restarted failed!", self.kernel_id)
-        self._send_status_message('dead')
+        self._send_status_message("dead")
 
     def _on_error(self, msg):
         if self.kernel_manager.allow_tracebacks:
             return
-        msg['content']['ename'] = 'ExecutionError'
-        msg['content']['evalue'] = 'Execution error'
-        msg['content']['traceback'] = [self.kernel_manager.traceback_replacement_message]
+        msg["content"]["ename"] = "ExecutionError"
+        msg["content"]["evalue"] = "Execution error"
+        msg["content"]["traceback"] = [self.kernel_manager.traceback_replacement_message]
 
-#-----------------------------------------------------------------------------
+
+# -----------------------------------------------------------------------------
 # URL to handler mappings
-#-----------------------------------------------------------------------------
+# -----------------------------------------------------------------------------
 
 
 _kernel_id_regex = r"(?P<kernel_id>\w+-\w+-\w+-\w+-\w+)"
